from rdkit import Chem
from rdkit.Chem import rdMolDescriptors
from novana.exception import NovanaError
from novana.exception import NonCyclicMoleculeError


class CyclicMoleculeBuilder:
    @staticmethod
    def create_rwmol_from_smiles(smiles, flatten_mol):
        """Flattens and creates an editable molecule."""
        rwmol = CyclicMoleculeBuilder._create_rwmol_from_smiles(smiles, flatten_mol)
        if rwmol is None:
            raise NovanaError("The molecule was not created for "
                              f"the input '{smiles}'")
        return rwmol

    @staticmethod
    def _create_rwmol_from_smiles(smiles, flatten_mol):
        """Business logic for rwmol creation."""
        if isinstance(smiles, str):
            if flatten_mol:
                smiles = CyclicMoleculeBuilder._remove_stereochemistry_from_smiles(
                    smiles)
            mol = Chem.MolFromSmiles(smiles)
            if mol is None:
                return None
            mol = CyclicMoleculeBuilder._keep_largest_fragment(mol)
            return Chem.RWMol(mol)

    @staticmethod
    def _remove_stereochemistry_from_smiles(smiles):
        return smiles.replace('@', '')

    @staticmethod
    def _keep_largest_fragment(mol):
        """
        Returns the largest fragment in a mixture.
        The fragment should have at least a ring,
        otherwise None will be returned.

        """
        mols = list(Chem.rdmolops.GetMolFrags(mol, asMols=True))
        if mols:
            mols.sort(reverse=True, key=lambda m: m.GetNumAtoms())
            for m in mols:
                if rdMolDescriptors.CalcNumRings(m) > 0:
                    return m
            # If no molecules with rings
            raise NonCyclicMoleculeError()



from novana.chem import CyclicMoleculeBuilder
from novana.process import TerminalAtomRemover
from novana.process import SingleBondedAtomConverter
from novana.process import SingleBondedCarbonConverter
from novana.process import HeteroAtomConverter
from novana.valence import HydrogenAdjuster
from novana.utils import get_logger
from novana.utils import disable_rdkit_logging


# Configure logging
log = get_logger(__name__)
disable_rdkit_logging()


def create_rwmol_from_smiles(smiles, flatten_mol=True):
    return CyclicMoleculeBuilder.create_rwmol_from_smiles(smiles, flatten_mol)


def adjust_hydrogens(rwmol):
    """Adjusts hydrogens in the molecule using heuristics."""
    return HydrogenAdjuster(rwmol).process_and_return()


def adjust_atom_hs_before_split(rwmol, bond):
    """Adjusts hydrogens in the molecule using heuristics."""
    return HydrogenAdjuster.adjust_atom_hs_before_split(rwmol, bond)


def recursively_remove_single_bonded_atoms(rwmol):
    """Removes recursively all single-bonded atoms and their bonds."""
    return TerminalAtomRemover(rwmol).process_and_return()


def convert_into_single_bonded_atoms(rwmol):
    """Converts all atoms into single-bonded carbons."""
    return SingleBondedAtomConverter(rwmol).process_and_return()


def convert_into_single_bonded_carbons(rwmol):
    """Converts all atoms into single-bonded carbons."""
    return SingleBondedCarbonConverter(rwmol).process_and_return()


def convert_heteroatoms_to_stars(rwmol):
    """Converts all heteroatoms into stars."""
    return HeteroAtomConverter(rwmol).process_and_return()



from novana.core import create_rwmol_from_smiles
from utils import mol_to_smiles


def test_largest_fragment_1():
    """The correct fragment is extracted from the mixture."""
    smiles = "N1CCNCC1.N2CCNCC2.N3CCNCC3.OC(CC(=O)O)(CC(=O)O)C(=O)O.OC(CC(=O)O)(CC(=O)O)C(=O)O"
    result = mol_to_smiles(create_rwmol_from_smiles(smiles))
    expected = "C1CNCCN1"
    assert result == expected


def test_largest_fragment_2():
    """The correct fragment is extracted from the mixture."""
    smiles = "C=C1CN(S(=O)(=O)c2ccc(C)cc2)CCCN(Cc2ccccc2)CCCN(S(=O)(=O)c2cccc(N(C)C)c2)C1.Cl"
    result = mol_to_smiles(create_rwmol_from_smiles(smiles))
    expected = "C=C1CN(S(=O)(=O)c2ccc(C)cc2)CCCN(Cc2ccccc2)CCCN(S(=O)(=O)c2cccc(N(C)C)c2)C1"
    assert result == expected


def test_largest_fragment_3():
    """The correct fragment is extracted from the mixture."""
    smiles = "[s+]1c3c(nc2c1cc(cc2)N(C)C)ccc(c3)N(C)C" "." \
             "[N+](=O)([O-])c4c(ccc(c4)[N+](=O)[O-])" \
             "NN=C(c5ccc(cc5)O)c6ccc(cc6)O"
    result = mol_to_smiles(create_rwmol_from_smiles(smiles))
    expected = "O=[N+]([O-])c1ccc(NN=C(c2ccc(O)cc2)c2ccc(O)cc2)c([N+](=O)[O-])c1"
    assert result == expected



import os
import logging
from rdkit import RDLogger


def get_logger(name):
    """
    Creates a logger where its level is
    determined by the variable NOVANA_LOG_LEVEL.

    """
    log = logging.getLogger(name)
    if os.environ.get("NOVANA_LOG_LEVEL", "").upper() == "DEBUG":
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)
    return log


def disable_rdkit_logging():
    """Disables INFO and WARNING logs from RDKit."""
    RDLogger.DisableLog('rdApp.info')
    RDLogger.DisableLog('rdApp.warning')



from novana.core import adjust_hydrogens
from novana.core import convert_into_single_bonded_carbons
from novana.core import convert_into_single_bonded_atoms
from novana.core import convert_heteroatoms_to_stars
from novana.core import create_rwmol_from_smiles
from novana.core import recursively_remove_single_bonded_atoms
from novana.enumerate import find_removable_groups
from novana.enumerate import enumerate_substructures


def substructure_scaffolds_from_smiles(smiles, flatten_mol=True):
    """Enumerates the partial scaffold decompositions of a molecule."""
    rwmol = create_rwmol_from_smiles(smiles, flatten_mol)
    removable_groups = find_removable_groups(rwmol)
    return enumerate_substructures(rwmol, removable_groups)


def scaffold_from_smiles(smiles, flatten_mol=True,
                         generalise_heteroatoms=False):
    """Returns the scaffold of a molecule."""
    rwmol = create_rwmol_from_smiles(smiles, flatten_mol)
    recursively_remove_single_bonded_atoms(rwmol)
    if generalise_heteroatoms:
        convert_heteroatoms_to_stars(rwmol)
    adjust_hydrogens(rwmol)
    return rwmol.GetMol()


def shape_from_smiles(smiles, flatten_mol=True,
                      retain_atom_types=False):
    """Returns the shape of a molecule."""
    rwmol = create_rwmol_from_smiles(smiles, flatten_mol)
    recursively_remove_single_bonded_atoms(rwmol)
    if retain_atom_types:
        convert_into_single_bonded_atoms(rwmol)
    else:
        convert_into_single_bonded_carbons(rwmol)
    return rwmol.GetMol()


def molecule_scaffold_shape_from_smiles(smiles,
                                        flatten_mol=True):
    """Returns molecule, scaffold, and shape of a molecule."""
    rwmol = create_rwmol_from_smiles(smiles, flatten_mol)
    mol = rwmol.GetMol()
    # Get scaffold and clean it
    recursively_remove_single_bonded_atoms(rwmol)
    adjust_hydrogens(rwmol)
    scaffold = rwmol.GetMol()

    # Convert scaffold into shape
    convert_into_single_bonded_carbons(rwmol)
    shape = rwmol.GetMol()
    return mol, scaffold, shape



from novana.api import substructure_scaffolds_from_smiles
from utils import mol_list_to_smiles_list


def test_enumerate_kekule():
    smiles = "ClC1=CC2=C(C=C1)N(C(=O)N2)C"
    result = mol_list_to_smiles_list(
        substructure_scaffolds_from_smiles(smiles))
    expected = ['Cn1c(=O)[nH]c2ccccc21', 'CN1CNc2cc(Cl)ccc21', 'O=c1[nH]c2ccc(Cl)cc2[nH]1']
    assert result == expected


def test_enumerate_remove_linker_product():
    smiles = "C(C)(C)COC1CC[C@H]2CN(C[C@H]2C1)C(=O)C[C@H](CC3=C(F)C=CC=C3Br)C(=O)O"
    result = mol_list_to_smiles_list(
        substructure_scaffolds_from_smiles(smiles))
    expected = ['O=C(O)C(CC(=O)N1CC2CCCCC2C1)Cc1c(F)cccc1Br',
                'CC(C)COC1CCC2CN(C(=O)CC(Cc3ccccc3F)C(=O)O)CC2C1',
                'CC(C)COC1CCC2CN(C(=O)CC(Cc3ccccc3Br)C(=O)O)CC2C1']
    assert result == expected



from novana.api import scaffold_from_smiles
from novana.api import shape_from_smiles
from utils import mol_to_inchikey
from utils import mol_to_smiles


def test_scaffold_1():
    """Scaffold decomposition."""
    smiles = "COC(=O)N1CC2(C1)CS(=O)(=Nc1cc(C)c3c(Nc4ccc(F)cc4O[C@H](C)C(=O)NCC(F)(F)F)ncnc3c1)C2"
    result = mol_to_smiles(scaffold_from_smiles(smiles))
    expected = "c1ccc(Nc2ncnc3cc(N=S4CC5(CNC5)C4)ccc23)cc1"
    assert result == expected


def test_scaffold_2():
    """Scaffold decomposition."""
    smiles = "P(=O)(OCc1ccc(cc1)OC(=O)c2ccccc2)(OCc3ccc(cc3)OC(=O)c4ccccc4)CCCC(=O)N(O)C"
    result = mol_to_smiles(scaffold_from_smiles(smiles))
    expected = "c1ccc(COc2ccc(COPOCc3ccc(OCc4ccccc4)cc3)cc2)cc1"
    assert result == expected


def test_scaffold_3():
    """Scaffold decomposition."""
    smiles = "C=C1CN(S(=O)(=O)c2ccc(C)cc2)CCCN(Cc2ccccc2)CCCN(S(=O)(=O)c2cccc(N(C)C)c2)C1"
    result = mol_to_smiles(scaffold_from_smiles(smiles))
    expected = "c1ccc(CN2CCCN(Sc3ccccc3)CCCN(Sc3ccccc3)CCC2)cc1"
    assert result == expected


def test_scaffold_4():
    """Scaffold decomposition."""
    smiles = "[O-]S(=O)(=O)C(O)[C@@H](NC(=O)[C@H](NC(=O)OC1(CCN(CC1)C(=O)OC(C)(C)C)CC)CC(C)C)C[C@@H]2CCNC2=O"
    result = mol_to_smiles(scaffold_from_smiles(smiles))
    expected = "C(CNCOC1CCNCC1)NCCC1CCNC1"
    assert result == expected


def test_scaffold_5():
    """Scaffold decomposition."""
    smiles = "C[C@@H]1CC[C@@]23CCC(=O)[C@H]2[C@@]1(C)[C@@H](C[C@@](C)(C=C)[C@@H](O)[C@@H]3C)OC=O"
    result = mol_to_smiles(scaffold_from_smiles(smiles))
    expected = "C1CCC2CCCC3(CC1)CCCC23"
    assert result == expected


def test_scaffold_6():
    """Scaffold decomposition."""
    smiles = r"CC1=CCC(O)/C=C\C(C)C(O)C(C)C=C(C)C(=O)c2c(O)c(C)cc3c2C(=O)C(N)=C(NC(=O)/C=C\C=C/C=C\C(C)C(O)CC1=O)C3=O"
    result = mol_to_smiles(scaffold_from_smiles(smiles))
    expected = r"C1=CCCCC/C=C\C=C/C=C\CNC2=CCc3c(cccc3C2)CC=CCCC/C=C\CC1"
    assert result == expected


def test_scaffold_7():
    """Scaffold decomposition."""
    smiles = "CCN(CC)[c+]1sc2c(s1)C(c1ccc(Cl)cc1)Oc1c(I)cc(I)cc1-2"
    result = mol_to_smiles(scaffold_from_smiles(smiles))
    expected = "c1ccc(C2Oc3ccccc3C3=C2SCS3)cc1"
    assert result == expected


def test_scaffold_hetero_1():
    """Scaffold decomposition."""
    smiles = "COC(=O)N1CC2(C1)CS(=O)(=Nc1cc(C)c3c(Nc4ccc(F)cc4O[C@H](C)C(=O)NCC(F)(F)F)ncnc3c1)C2"
    result = mol_to_smiles(scaffold_from_smiles(smiles,
                           generalise_heteroatoms=True))
    expected = "*1CC2(C1)C*(=*c1ccc3c(*c4ccccc4)*c*c3c1)C2"
    assert result == expected


def test_scaffold_hetero_2():
    """Scaffold decomposition."""
    smiles = "P(=O)(OCc1ccc(cc1)OC(=O)c2ccccc2)(OCc3ccc(cc3)OC(=O)c4ccccc4)CCCC(=O)N(O)C"
    result = mol_to_smiles(scaffold_from_smiles(smiles,
                           generalise_heteroatoms=True))
    expected = "*(*Cc1ccc(*Cc2ccccc2)cc1)*Cc1ccc(*Cc2ccccc2)cc1"
    assert result == expected


def test_scaffold_hetero_3():
    """Scaffold decomposition."""
    smiles = "C=C1CN(S(=O)(=O)c2ccc(C)cc2)CCCN(Cc2ccccc2)CCCN(S(=O)(=O)c2cccc(N(C)C)c2)C1"
    result = mol_to_smiles(scaffold_from_smiles(smiles,
                           generalise_heteroatoms=True))
    expected = "*(*1CCC*(*c2ccccc2)CCC*(Cc2ccccc2)CCC1)c1ccccc1"
    assert result == expected


def test_scaffold_hetero_4():
    """Scaffold decomposition."""
    smiles = "[O-]S(=O)(=O)C(O)[C@@H](NC(=O)[C@H](NC(=O)OC1(CCN(CC1)C(=O)OC(C)(C)C)CC)CC(C)C)C[C@@H]2CCNC2=O"
    result = mol_to_smiles(scaffold_from_smiles(smiles,
                           generalise_heteroatoms=True))
    expected = "*(C*C1CC*CC1)CC*CCC1C*CC1"
    assert result == expected


def test_scaffold_hetero_5():
    """Scaffold decomposition."""
    smiles = "C[C@@H]1CC[C@@]23CCC(=O)[C@H]2[C@@]1(C)[C@@H](C[C@@](C)(C=C)[C@@H](O)[C@@H]3C)OC=O"
    result = mol_to_smiles(scaffold_from_smiles(smiles,
                           generalise_heteroatoms=True))
    expected = "C1CCC2CCCC3(CC1)CCCC23"
    assert result == expected


def test_scaffold_hetero_6():
    """Scaffold decomposition."""
    smiles = r"CC1=CCC(O)/C=C\C(C)C(O)C(C)C=C(C)C(=O)c2c(O)c(C)cc3c2C(=O)C(N)=C(NC(=O)/C=C\C=C/C=C\C(C)C(O)CC1=O)C3=O"
    result = mol_to_smiles(scaffold_from_smiles(smiles,
                           generalise_heteroatoms=True))
    expected = r"*1C/C=C\C=C/C=C\CCCCC=CCC/C=C\CCCC=CCc2cccc3c2CC=C1C3"
    assert result == expected


def test_scaffold_hetero_7():
    """Scaffold decomposition."""
    smiles = "CCN(CC)[c+]1sc2c(s1)C(c1ccc(Cl)cc1)Oc1c(I)cc(I)cc1-2"
    result = mol_to_smiles(scaffold_from_smiles(smiles,
                           generalise_heteroatoms=True))
    expected = "*1c2ccccc2C2=*C*=C2C1c1ccccc1"
    assert result == expected


def test_scaffold_mixture_1():
    """Scaffold decomposition."""
    smiles = "S2c1c(cc(cc1)C(=O)C)N(c3c2cccc3)CCCN4CCN(CC4)CCO" "." \
             r"OC(=O)/C=C\C(=O)O" "." r"OC(=O)/C=C\C(=O)O"
    result = mol_to_smiles(scaffold_from_smiles(smiles))
    expected = "c1ccc2c(c1)Sc1ccccc1N2CCCN1CCNCC1"
    assert result == expected


def test_scaffold_mixture_2():
    """Scaffold decomposition."""
    smiles = "C=C1CN(S(=O)(=O)c2ccc(C)cc2)CCCN(Cc2ccccc2)" \
             "CCCN(S(=O)(=O)c2cccc(N(C)C)c2)C1" "." "Cl"
    result = mol_to_smiles(scaffold_from_smiles(smiles))
    expected = "c1ccc(CN2CCCN(Sc3ccccc3)CCCN(Sc3ccccc3)CCC2)cc1"
    assert result == expected


def test_scaffold_mixture_3():
    """Scaffold decomposition."""
    smiles = "[Na+].[O-]S(=O)(=O)C(O)[C@@H](NC(=O)[C@H](NC(=O)" \
             "OC1(CCN(CC1)C(=O)OC(C)(C)C)CC)CC(C)C)C[C@@H]2CCNC2=O"
    result = mol_to_smiles(scaffold_from_smiles(smiles))
    expected = "C(CNCOC1CCNCC1)NCCC1CCNC1"
    assert result == expected


def test_scaffold_mixture_4():
    """Scaffold decomposition."""
    smiles = "[I-].CC[n+]1ccccc1CNC2cccc2"
    result = mol_to_smiles(scaffold_from_smiles(smiles))
    expected = "C1=CC(NCc2ccccn2)C=C1"
    assert result == expected


def test_scaffold_charged_1():
    """Scaffold decomposition."""
    smiles = "[BH2-]1n2cccc2C(c2ccccc2)=C2C=CC=[N+]12"
    result = mol_to_smiles(scaffold_from_smiles(smiles))
    expected = "B1n2cccc2C(c2ccccc2)=C2C=CC=[N+]12"
    assert result == expected


def test_scaffold_charged_2():
    """Scaffold decomposition."""
    smiles = "CC[n+]1ccccc1CNC2cccc2"
    result = mol_to_smiles(scaffold_from_smiles(smiles))
    expected = "C1=CC(NCc2ccccn2)C=C1"
    assert result == expected


def test_scaffold_charged_3():
    """Scaffold decomposition."""
    smiles = "CCN(CC)[c+]1sccs1"
    result = mol_to_smiles(scaffold_from_smiles(smiles))
    expected = "C1=CSCS1"
    assert result == expected


def test_scaffold_charged_4():
    """Scaffold decomposition."""
    smiles = "C[C@H]1[C@H]2[C@H](C[C@H]3[C@@H]4CC=C5C[C@H](n6cc[n+](Cc7ccccc7Br)c6)CC[C@]5(C)[C@H]4CC[C@@]32C)O[C@]12CC[C@@H](C)CO2"  # noqa: E501
    result = mol_to_smiles(scaffold_from_smiles(smiles))
    expected = "C1=C2CC(n3cc[n+](Cc4ccccc4)c3)" \
               "CCC2C2CCC3C4CC5(CCCCO5)OC4CC3C2C1"
    assert result == expected


def test_scaffold_charged_5():
    """Scaffold decomposition."""
    smiles = "[BH2-]1n2cccc2C(c2ccccc2)=C2C=CC=[N+]12"
    result = mol_to_smiles(scaffold_from_smiles(smiles))
    expected = "B1n2cccc2C(c2ccccc2)=C2C=CC=[N+]12"
    assert result == expected


def test_shape_1():
    """Shape decomposition."""
    smiles = "CCN(CC)[c+]1sc2c(s1)C(c1ccc(Cl)cc1)Oc1c(I)cc(I)cc1-2"
    result = mol_to_smiles(shape_from_smiles(smiles))
    expected = "C1CC2C3CCCCC3CC(C3CCCCC3)C2C1"
    assert result == expected


def test_shape_2():
    """Shape decomposition."""
    smiles = "C[C@@H]1CC[C@@]23CCC(=O)[C@H]2[C@@]1(C)[C@@H](C[C@@](C)(C=C)[C@@H](O)[C@@H]3C)OC=O"
    result = mol_to_smiles(shape_from_smiles(smiles))
    expected = "C1CCCC2CCCC3(C1)CCCC23"
    assert result == expected


def test_shape_3():
    """Shape decomposition."""
    smiles = r"CC1=CCC(O)/C=C\C(C)C(O)C(C)C=C(C)C(=O)c2c(O)c(C)cc3c2C(=O)C(N)=C(NC(=O)/C=C\C=C/C=C\C(C)C(O)CC1=O)C3=O"
    result = mol_to_smiles(shape_from_smiles(smiles))
    expected = "C1CCCCCCCCCCCCC2CCC3C(CCCCCCCCCCC1)CCCC3C2"
    assert result == expected


def test_shape_types_1():
    """Shape decomposition."""
    smiles = "CCN(CC)[c+]1sc2c(s1)C(c1ccc(Cl)cc1)Oc1c(I)cc(I)cc1-2"
    result = mol_to_smiles(shape_from_smiles(smiles,
                                             retain_atom_types=True))
    expected = "C1SC2C3CCCCC3OC(C3CCCCC3)C2S1"
    assert result == expected


def test_shape_types_2():
    """Shape decomposition."""
    smiles = "C[C@@H]1CC[C@@]23CCC(=O)[C@H]2[C@@]1(C)[C@@H](C[C@@](C)(C=C)[C@@H](O)[C@@H]3C)OC=O"
    result = mol_to_smiles(shape_from_smiles(smiles,
                                             retain_atom_types=True))
    expected = "C1CCCC2CCCC3(C1)CCCC23"
    assert result == expected


def test_shape_types_3():
    """Shape decomposition."""
    smiles = r"CC1=CCC(O)/C=C\C(C)C(O)C(C)C=C(C)C(=O)c2c(O)c(C)cc3c2C(=O)C(N)=C(NC(=O)/C=C\C=C/C=C\C(C)C(O)CC1=O)C3=O"
    result = mol_to_smiles(shape_from_smiles(smiles,
                                             retain_atom_types=True))
    expected = "C1CCCCCCCCCCCNC2CCC3C(CCCCCCCCCCC1)CCCC3C2"
    assert result == expected


def test_identical_scaffold_shape_1():
    """Scaffold and shape are identical."""
    smiles = "C[C@@H]1CC[C@@]23CCC(=O)[C@H]2[C@@]1(C)[C@@H](C[C@@](C)(C=C)[C@@H](O)[C@@H]3C)OC=O"
    scaffold = mol_to_inchikey(scaffold_from_smiles(smiles))
    shape = mol_to_inchikey(shape_from_smiles(smiles))
    assert scaffold == shape


def test_different_scaffold_tautomers_1():
    """Tautomers yield different scaffolds."""
    smiles = r"CC1=CCC(O)/C=C\C(C)C(O)C(C)C=C(C)C(=O)c2c(O)c(C)cc3c2C(=O)C(N)=C(NC(=O)/C=C\C=C/C=C\C(C)C(O)CC1=O)C3=O"
    result_1 = mol_to_inchikey(scaffold_from_smiles(smiles))
    smiles = r"CC1\C=C/C=C\C=C/C(=NC2=C(N)C(=O)c3c(cc(C)c(O)c3C(=O)C(=CC(C)C(O)C(C)\C=C/C(O)CC=C(C)C(=O)CC1O)C)C2=O)O"
    result_2 = mol_to_inchikey(scaffold_from_smiles(smiles))
    assert result_1 != result_2


def test_identical_shape_tautomers_1():
    """Tautomers yield same shape."""
    smiles = r"CC1=CCC(O)/C=C\C(C)C(O)C(C)C=C(C)C(=O)c2c(O)c(C)cc3c2C(=O)C(N)=C(NC(=O)/C=C\C=C/C=C\C(C)C(O)CC1=O)C3=O"
    result_1 = mol_to_inchikey(shape_from_smiles(smiles))
    smiles = r"CC1\C=C/C=C\C=C/C(=NC2=C(N)C(=O)c3c(cc(C)c(O)c3C(=O)C(=CC(C)C(O)C(C)\C=C/C(O)CC=C(C)C(=O)CC1O)C)C2=O)O"
    result_2 = mol_to_inchikey(shape_from_smiles(smiles))
    assert result_1 == result_2



from novana.core import adjust_hydrogens
from novana.core import adjust_atom_hs_before_split
from novana.process import TerminalAtomRemover
from copy import deepcopy
from rdkit import Chem
from rdkit.Chem import rdmolops


def find_removable_groups(rwmol):
    initial_groups = _find_terminal_atom_groups(rwmol)
    expanded_groups = _expand_atom_groups(rwmol, initial_groups)
    return deduplicate_unsorted_list_of_lists(expanded_groups)


def _find_terminal_atom_groups(rwmol):
    """Create a list of singletons of terminal atoms in the molecule."""
    terminal_atoms_groups = []
    for bond in list(rwmol.GetBonds()):
        for atom in (bond.GetBeginAtom(), bond.GetEndAtom()):
            if TerminalAtomRemover._is_atom_single_bonded(atom):
                terminal_atoms_groups.append([atom.GetIdx()])
    return terminal_atoms_groups


def _expand_atom_groups(rwmol, atoms_groups):
    for group in atoms_groups:
        for atom_idx in group:
            atom = rwmol.GetAtomWithIdx(atom_idx)
            bonds = atom.GetBonds()
            for b in bonds:
                neighbour_atom = _get_neighbour_atom(b, atom_idx)
                if neighbour_atom.IsInRing():
                    continue
                neighbour_atom_idx = neighbour_atom.GetIdx()
                if neighbour_atom_idx not in group:
                    group.append(neighbour_atom_idx)
    return atoms_groups


def _get_neighbour_atom(bond, query_atom_idx):
    if bond.GetBeginAtom().GetIdx() == query_atom_idx:
        return bond.GetEndAtom()
    else:
        return bond.GetBeginAtom()


def deduplicate_unsorted_list_of_lists(list_of_lists):
    return [list(x) for x in set(tuple(sorted(lst)) for lst in list_of_lists)]


def remove_isolated_atoms(rwmol):
    isolated_atoms = [atom.GetIdx() for atom in rwmol.GetAtoms() if atom.GetDegree() == 0]
    for idx in sorted(isolated_atoms, reverse=True):
        rwmol.RemoveAtom(idx)
    return rwmol


def enumerate_substructures(rwmol, removable_groups):
    """Enumerates the partial decompositions of a molecule."""
    enumerated_mols = list()
    for group in removable_groups:
        rw_substructure = create_molecule_with_removed_group(rwmol, group)
        # If decomposition has created multiple fragments, skip it
        if not _has_molecule_multiple_fragments(rw_substructure):
            enumerated_mols.append(rw_substructure)
    return enumerated_mols


def create_molecule_with_removed_group(rwmol, group):
    """Creates a new molecule with the specified group removed."""
    _rwmol = deepcopy(rwmol)
    _rwmol = _delete_group_from_mol(_rwmol, group)
    _rwmol = adjust_hydrogens(_rwmol)
    Chem.SanitizeMol(_rwmol)
    return _rwmol


def _delete_group_from_mol(rwmol, group):
    """Delete a group of atoms and their bonds from a molecule."""
    reversed_group = sorted(group, reverse=True)  # To avoid index shifting
    for atom_idx in reversed_group:
        bonds = rwmol.GetAtomWithIdx(atom_idx).GetBonds()
        for bond in bonds:
            neighbour_atom = _get_neighbour_atom(bond, atom_idx)
            adjust_atom_hs_before_split(neighbour_atom, bond)
            rwmol.RemoveBond(atom_idx,
                             neighbour_atom.GetIdx())
    return remove_isolated_atoms(rwmol)


def _has_molecule_multiple_fragments(mol):
    """Check if a molecule has multiple fragments."""
    fragment_ids = rdmolops.GetMolFrags(mol)
    return len(fragment_ids) > 1



from novana.valence import HydrogenAdjuster
from rdkit import Chem


class TerminalAtomRemover(object):
    def __init__(self, rwmol):
        self.mol = rwmol

    def process_and_return(self):
        self.remove_atoms_recursively()
        return self.mol

    def remove_atoms_recursively(self):
        """
        Removes recursively all terminal
        atoms and their bonds.

        """
        init_atom_length = len(self.mol.GetAtoms())
        final_atom_length = TerminalAtomRemover._set_dummy_length()
        while init_atom_length != final_atom_length:
            init_atom_length = len(self.mol.GetAtoms())
            self._remove_atoms()
            final_atom_length = len(self.mol.GetAtoms())
        Chem.SanitizeMol(self.mol)
        Chem.rdmolops.SanitizeFlags.SANITIZE_NONE

    @staticmethod
    def _set_dummy_length():
        return 0  # dummy length to start the while loop

    def _remove_atoms(self):
        """
        Iterates through the molecule's atoms and
        removes all atoms that are terminal (and
        their bonds).
        """
        for bond in list(self.mol.GetBonds()):
            for atom in (bond.GetBeginAtom(), bond.GetEndAtom()):
                if TerminalAtomRemover._is_atom_single_bonded(atom):
                    nbr_atom = bond.GetOtherAtom(atom)
                    if TerminalAtomRemover._is_atom_multi_bonded(nbr_atom):
                        self._remove_single_bonded_atom_from_mol(
                            single_bonded_atom=atom,
                            neighbour_atom=nbr_atom,
                            bond=bond)
                        break

    @staticmethod
    def _is_atom_single_bonded(atom):
        return atom.GetDegree() == 1

    @staticmethod
    def _is_atom_multi_bonded(atom):
        return atom.GetDegree() > 1

    def _remove_single_bonded_atom_from_mol(self, single_bonded_atom,
                                            neighbour_atom, bond):
        HydrogenAdjuster.adjust_atom_hs_before_split(
            neighbour_atom, bond)
        self.mol.RemoveBond(single_bonded_atom.GetIdx(),
                            neighbour_atom.GetIdx())
        self.mol.RemoveAtom(single_bonded_atom.GetIdx())


class AtomConverter:
    """Interface for specialised converters."""
    def __init__(self, rwmol):
        self.mol = rwmol

    def process_and_return(self):
        self.convert()
        return self.mol

    def convert(self):
        pass


class SingleBondedAtomConverter(AtomConverter):
    def convert(self):
        """Converts all atoms into single-bonded."""
        for bond in list(self.mol.GetBonds()):
            bond.SetBondType(Chem.BondType.SINGLE)
            bond.SetIsAromatic(False)
            for atom in (bond.GetBeginAtom(), bond.GetEndAtom()):
                self._change_properties(atom)
                break

    def _change_properties(self, atom):
        atom.SetIsAromatic(False)
        atom.SetFormalCharge(0)
        atom.SetNumExplicitHs(0)


class SingleBondedCarbonConverter(SingleBondedAtomConverter):
    def _change_properties(cls, atom):
        atom.SetIsAromatic(False)
        atom.SetFormalCharge(0)
        atom.SetNumExplicitHs(0)
        atom.SetAtomicNum(6)


class HeteroAtomConverter(AtomConverter):
    NON_HETERO_ATOM_NUMS = [
        Chem.Atom("H").GetAtomicNum(),
        Chem.Atom("C").GetAtomicNum()]

    def convert(self):
        """Converts hetero atoms into stars."""
        for bond in list(self.mol.GetBonds()):
            for atom in (bond.GetBeginAtom(), bond.GetEndAtom()):
                self._change_properties(atom)
                break

    def _change_properties(self, atom):
        if self._is_hetero_atom(atom):
            atom.SetAtomicNum(0)
            atom.SetNumExplicitHs(0)
            atom.UpdatePropertyCache()

    def _is_hetero_atom(self, atom):
        if atom.GetAtomicNum() not in \
                HeteroAtomConverter.NON_HETERO_ATOM_NUMS:
            return True
        return False



class NovanaError(Exception):
    pass


class NonCyclicMoleculeError(Exception):
    def __init__(self, message="The input does not contain "
                               "any cyclic molecule."):
        super().__init__(message)



from math import ceil
from rdkit import Chem
from novana.utils import get_logger
from novana.utils import disable_rdkit_logging


# Configure logging
log = get_logger(__name__)
disable_rdkit_logging()

# Static objects
periodic_table = Chem.GetPeriodicTable()


class AtomProperties(object):
    def __init__(self, atom):
        self.atom = atom
        self.symbol = atom.GetSymbol()
        self.index = atom.GetIdx()
        self.atomic_num = atom.GetAtomicNum()
        self.explict_hs_num = atom.GetNumExplicitHs()
        self.total_valence = atom.GetTotalValence()
        self.formal_charge = atom.GetFormalCharge()
        self.default_valence = self._calculate_default_valence(periodic_table)

    def _calculate_default_valence(self, periodic_table):
        return periodic_table.GetDefaultValence(self.atomic_num)

    def are_there_explicit_hs(self):
        if self.explict_hs_num > 0:
            return True
        return False

    def log_debug_properties(self):
        log.debug(f"Atom: {self.symbol}")
        log.debug(f"Atom index: {self.index}")
        log.debug(f"Charge: {self.formal_charge}")
        log.debug(f"Total valence: {self.total_valence}")
        log.debug(f"Default valence: {self.default_valence}")
        log.debug(f"Explicit hydrogens {self.explict_hs_num}")


class HydrogenAdjuster(object):
    def __init__(self, mol):
        self.mol = mol

    def process_and_return(self):
        self.run()
        return self.mol

    def run(self):
        """
        Sanitisation heuristics to fix hypervalent atoms
        by removing hydrogens in excess; and hypovalent
        atoms by adding hydrogens. Both processes are done
        in order to achieve the default valence of the atom.

        """
        log.debug(Chem.MolToSmiles(self.mol))
        for bond in list(self.mol.GetBonds()):
            for atom in (bond.GetBeginAtom(), bond.GetEndAtom()):
                atom_props = AtomProperties(atom)
                if atom_props.are_there_explicit_hs:
                    bond_type = bond.GetBondTypeAsDouble()
                    atom_props.log_debug_properties()
                    log.debug(f"Bond type as double: {bond_type}")

                    # If total valence is greater than default, hydrogens
                    # need to be removed and charges neutralised
                    if atom_props.total_valence > atom_props.default_valence:
                        HydrogenAdjuster.remove_hydrogens(atom, atom_props,
                                                          bond_type)

                    # If total valence is smaller than default, hydrogens
                    # need to be added and charges neutralised
                    # In addition, if aromatic bonds are present,
                    # they should be converted into single to accommodate
                    # the previous changes, otherwise kekulisation will fail
                    elif atom_props.total_valence < atom_props.default_valence:
                        self.add_hydrogens(atom, atom_props, bond, bond_type)

                    atom.UpdatePropertyCache()
                    break

    @staticmethod
    def remove_hydrogens(atom, atom_props, bond_type):
        hs_to_remove = int(
            ceil((atom_props.total_valence - atom_props.default_valence)
                 / bond_type))
        if hs_to_remove <= atom_props.explict_hs_num and hs_to_remove > 0:
            atom.SetNumExplicitHs(atom_props.explict_hs_num - hs_to_remove)
            atom.SetFormalCharge(0)

    def add_hydrogens(self, atom, atom_props, bond, bond_type):
        if bond_type == 1.5:
            bond.SetBondType(Chem.rdchem.BondType.SINGLE)
        atom.SetNumExplicitHs(atom_props.explict_hs_num + 1)
        atom.SetFormalCharge(0)
        Chem.SanitizeMol(self.mol)
        Chem.rdmolops.SanitizeFlags.SANITIZE_NONE

    @staticmethod
    def adjust_atom_hs_before_split(atom, bond):
        """
        Increase the explicit hydrogens of an
        atom by the rounded value of its bond type.

        """
        atom.SetNumExplicitHs(
            atom.GetNumExplicitHs()
            + int(bond.GetBondTypeAsDouble()))



from setuptools import find_packages, setup

__docs__ = "NovAna (Novelty Analysis) is a cheminformatics tool that " \
           "that extends the decomposition of molecules into " \
           "Bemis-Murcko frameworks (J. Med. Chem. 1996, 39, 15, 2887-2893) " \
           "and allows the enumeration of partially decomposed scaffolds."

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="novana",
    packages=find_packages("src"),
    version="0.3.0",
    description=__docs__,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Gian Marco Ghiandoni",
    author_email="ghiandoni.g@gmail.com",
    license="MIT",
    install_requires=["rdkit"],
    setup_requires=["pytest-runner", "flake8"],
    tests_require=["pytest==7.1.2"],
    test_suite="test",
    package_dir={"novana": "src/novana"},
)



import pytest
from novana.api import scaffold_from_smiles
from novana.api import shape_from_smiles
from novana.exception import NovanaError
from novana.exception import NonCyclicMoleculeError


def test_exception_invalid_input_scaffold_1():
    """Raises an error if the input is invalid."""
    smiles = "AAA"
    with pytest.raises(NovanaError):
        scaffold_from_smiles(smiles)


def test_exception_invalid_input_shape_1():
    """Raises an error if the input is invalid."""
    smiles = "AAA"
    with pytest.raises(NovanaError):
        shape_from_smiles(smiles)


def test_exception_invalid_input_scaffold_2():
    """Raises an error if the input is invalid."""
    smiles = 0
    with pytest.raises(NovanaError):
        scaffold_from_smiles(smiles)


def test_exception_invalid_input_shape_2():
    """Raises an error if the input is invalid."""
    smiles = 0
    with pytest.raises(NovanaError):
        shape_from_smiles(smiles)


def test_exception_mixture_no_rings_scaffold():
    """Raises an error if no molecules with rings."""
    smiles = "OC(CC(=O)O)(CC(=O)O)C(=O)O.OC(CC(=O)O)(CC(=O)O)C(=O)O"
    with pytest.raises(NonCyclicMoleculeError):
        scaffold_from_smiles(smiles)


def test_exception_molecule_no_rings_scaffold():
    """Raises an error if no molecules with rings."""
    smiles = "CCCCCCCCCCNC(=O)C[N+](C)(CCO)CCCCCCCCCC[N+](C)(CCO)CC(=O)NCCCCCCCCCC"
    with pytest.raises(NonCyclicMoleculeError):
        scaffold_from_smiles(smiles)


def test_exception_mixture_no_rings_shape():
    """Raises an error if no molecules with rings."""
    smiles = "OC(CC(=O)O)(CC(=O)O)C(=O)O.OC(CC(=O)O)(CC(=O)O)C(=O)O"
    with pytest.raises(NonCyclicMoleculeError):
        shape_from_smiles(smiles)


def test_exception_molecule_no_rings_shape():
    """Raises an error if no molecules with rings."""
    smiles = "CCCCCCCCCCNC(=O)C[N+](C)(CCO)CCCCCCCCCC[N+](C)(CCO)CC(=O)NCCCCCCCCCC"
    with pytest.raises(NonCyclicMoleculeError):
        shape_from_smiles(smiles)



import time
from novana.api import scaffold_from_smiles
from novana.api import shape_from_smiles


def time_function(func, *args, **kwargs):
    """Returns the time (ms) needed to run a function."""
    start = time.time()
    func(*args, **kwargs)
    end = time.time()
    return (end - start)*1000


def test_time_scaffold_2_rings():
    """Scaffold decomposition within a time cap."""
    smiles = "CC[n+]1ccccc1CNC2cccc2"
    t = time_function(scaffold_from_smiles, smiles)
    assert t < 2


def test_time_scaffold_4_rings():
    """Scaffold decomposition within a time cap."""
    smiles = "S2c1c(cc(cc1)C(=O)C)N(c3c2cccc3)CCCN4CCN(CC4)CCO"
    t = time_function(scaffold_from_smiles, smiles)
    assert t < 3


def test_time_scaffold_8_rings():
    """Scaffold decomposition within a time cap."""
    smiles = "CC[n+]1ccccc1CNC2cccc2"*4
    t = time_function(scaffold_from_smiles, smiles)
    assert t < 4


def test_time_scaffold_20_rings():
    """Scaffold decomposition within a time cap."""
    smiles = "CC[n+]1ccccc1CNC2cccc2"*10
    t = time_function(scaffold_from_smiles, smiles)
    assert t < 8


def test_time_shape_2_rings():
    """Shape decomposition within a time cap."""
    smiles = "CC[n+]1ccccc1CNC2cccc2"
    t = time_function(shape_from_smiles, smiles)
    assert t < 2


def test_time_shape_4_rings():
    """Shape decomposition within a time cap."""
    smiles = "S2c1c(cc(cc1)C(=O)C)N(c3c2cccc3)CCCN4CCN(CC4)CCO"
    t = time_function(shape_from_smiles, smiles)
    assert t < 3


def test_time_shape_8_rings():
    """Shape decomposition within a time cap."""
    smiles = "CC[n+]1ccccc1CNC2cccc2"*4
    t = time_function(shape_from_smiles, smiles)
    assert t < 4


def test_time_shape_20_rings():
    """Shape decomposition within a time cap."""
    smiles = "CC[n+]1ccccc1CNC2cccc2"*10
    t = time_function(shape_from_smiles, smiles)
    assert t < 8
